#!/usr/bin/env python

import sys, os, commands, string

# these are used for error reporting.

filename=''
lineno=0

brackstart = '[['
brackend   = ']]'
default_rules = "MACROS.py"
version = '0.1'

################################################
# Common/share functions

def error(msg=''):
    """ Big grand error reporter. This is passed to env{} so anyone can use it."""

    print "ERROR: %s" % msg
    if lineno:
        print "File: %s, line %s" % (filename, lineno)
    sys.exit(1)


def usage():
    """ Display a nice usage message. """

    print "htmlexp: expands macros in source file for html output."
    print "Usage:  htmlexp <infile> <options>"
    print "Options:"
    print "   -r  <file> Read additional rule file"
    print "   -w  prohibit overwriting existing file."
    print "   -o <file>  set output filename"
    print "   -b <startbracket endbracket> Set macro delims"
    print "   -d  Prints macro defs. Note, uses << >> for delimiters."
    print "   -v  prints program version"
    print
    sys.exit(1)

def opts():
    """ Parse off command line options.

        Returns: input fileaname, file object for output. 
    """

    import getopt

    rulesdone = 0
    overwrite = 1
    outf = ''
    printdoc = 0

    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:],
            "r:o:wb:dv", [] )
    except getopt.GetoptError:
        usage()

    for o,a in opts:
        if o == '-r':
            execfile(a, env, env)
            rulesdone=1

        elif o == '-o':
            outf = a

        elif o == '-w':    # no overwrite
            overwrite = 0

        elif o == '-b':    # set delims
            parse("BRACKETS %s" % a)

        elif o == '-d':  # print macro documentation
            printdoc = 1

        elif o == '-v':
            print version
            sys.exit(0)

        else:
            usage()

    if not printdoc:
        if len(args) != 1:
            print "htmlexp: requires 1 filename argument."
            usage()
        else:
            fn = args[0]

    if not rulesdone:   # load default rules file
        if not os.path.exists(default_rules):
            print("Warning: No rules file imported.")
        execfile(default_rules, env, env)

    if printdoc:
        printdocs()
        sys.exit(0)

    if not outf: # The output filename is input minus extension plus ".html"
        outf = os.path.splitext(fn)[0] + ".html"

    # Overwrite test

    if not overwrite and os.path.exists(outf):
        error("File '%s' exists." % outf)

    # make filename available to functions

    env['HTMLFILE'] = outf
    env['FILENAME'] = fn

    outpath = file(outf, 'w')   # open the output now

    return (fn, outpath)

def printdocs():
    """ Print docs for what we think are user macros and variables. These are
        defined in the env{} and are in UPPERCASE.
    """

    print "<< SEC Functions >>"
    for a in sorted(env):
        if type(env[a]) == type(printdocs) and a.upper() == a:   # only all uppercase is used.
            s=env[a].__doc__
            if s:
                s = s.replace("<", "&lt;")
                s = s.replace(">", "&gt;")
            else:
                s = "Undocumented"
            print "<< FUNC %s {%s} >>" % (a, s)

    print "<< SEC Variables >>"
    for a in sorted(env):
        if type(env[a]) == type('a') and a.upper() == a:   # only all uppercase is used.
            print "<< VAR %s  {%s} >>" % ( a, env[a] )

####################################################
### Basic functions which are needed to do anything much.
### They can be redefined in a user script. After defining
### they are set in the env{} namespace.

def BRACKETS(p):
    """Redefine the brackets used to define a macro.

       args - <start bracket> <end bracket>
    """

    global brackstart, brackend

    if len(p) != 2:
        error("BRACKETS: set macro brackets <start> <end>")

    brackstart = p[0]
    brackend   = p[1]

    if len(brackstart) != len(brackend):
        error("BRACKETS: start/end should be same size.")

    if brackstart == brackend:
        error("BRACKETS: <start> cannot be the same as <end>.")

    return ''
    
def PFILE(p):
    """Import a macro file. Must contain Python functions and variables.

       args - <filename>
    """

    if len(p) != 1:
        error("PFILE: <filename>.")

    execfile(p[0], env, env)
    return ''

def PYTHON(p):
    """Call Python.  Can be used to define functions on-the-fly. 
       
       Note: all the python code should be passed in a single arg,
             so don't forget to bracket or use '#PYTHON'. Example:

             [ PYTHON '
             def FUNC(p):
                return "whatever" ' ] 
       
       Note: it is important to respect Python's indentation rules!!!

       args - <python code>

    """

    if len(p) != 1:
        error("PYTHON: <code>")
    
    exec p[0] in env, env
    return ''

def INCLUDE(p):
    """Include a source file or files. 

       args - <filename> [filename] ...
    """

    x = []
    if not len(p):
        error("INCLUDE: Expecting filename[s].")

    for f in p:
        x.extend ( dofile(f) )

    return ''.join(x)

def SETROOT(p):
    """Set the root directory. Value accessed with the [ROOTDIR] command.

       args - <dirname> 
    """

    global ROOTDIR

    if len(p) != 1:
        error("SETROOT: one arg required.")

    env['ROOTDIR'] = p[0]

    return ''

def DEFINE(p):
    """ Define a variable. If you just pass a single arg then the
        variable will be defined with a EMPTY value. """

    if not len(p):
        error("DEFINE: at least one arg is required.")

    env[p[0]] = ' '.join(p[1:])
    return ''

def ECHO(p):
    """ Print a variable(s). Variables not defined are silently ignored. 
        All variables are expanded and the result in concatenated. """

    ret = []
    for a in p:
        try: ret.append(env[a])
        except: pass
    
    return ' '.join(ret)

#### Set up our initial namespace.

env = {'PFILE':    PFILE ,
       'PYTHON':   PYTHON,
       'INCLUDE':  INCLUDE,
       'BRACKETS': BRACKETS,
       'SETROOT':  SETROOT,
       'ROOTDIR':  '.',
       'error':    error,
       'DEFINE':   DEFINE,
       'ECHO':     ECHO }

env['env'] =      env


##############################################################
### Parsing code

def lsplit(s):
    """ Split a string. """

    white=string.whitespace

    def setterm(c):
        if c=='"':     return  '"'
        elif c == "'": return  "'"
        elif c == '{': return  "}"
        else:          return  white

    s=s.strip()    # delete trailing/leading blanks

    if s[0] == '#':    # Handle "no interpret" args, eg [ #PYTHON ... ]
        s=s.split(None, 1)
        s[0]=s[0][1:].strip()
        return s

    r = []         # our split list
    chunk = []     # used to build items
    p=0            # pointer to current pos
    last=len(s)-1  # End of line

    term = setterm(s[0])  # initial chunk terminator

    if term not in white:       # skip over if NOT blank.
        p+=1

    while 1:              # main loop ... seek and chop
        c = s[p]
        if c in term:
            r.append(''.join(chunk))   # add chunk to the new list
            chunk = []
            
            # Advance to next position. Skip duplicate spaces

            p+=1
            while p<last and s[p] in white:
                p+=1

            # if at EOL it's okay to exit.

            if p > last:
                break

            # Get new chunk terminator 

            term = setterm(s[p])

            if term != white:   # skip terminator
                p+=1

        else:             # build up this chunk
            chunk.append(c)
            p += 1
            
        if p>last:        # check for eol
            if term in white:
                r.append(''.join(chunk))
                break
            else:
                error("EOL reached searching for '%s' in'%s'." % (term, s))

    return r


            
def parse(s):
    """ Expand a command which was set in {}s. """

    s = lsplit(s)

    if not len(s):
        error("Expecting something in [] command")

    cmd = s[0]
    args = s[1:]

    if not cmd in env:
        error("Unknown function/variable '%s'" % cmd)
    
    if type(env[cmd]) == type(parse):
        txt = env[cmd](args)

    else:   # not a func, just expand the variable
        if args:
            error("Argument to variable '%s' not permitted." % cmd)
        txt = str(env[cmd])

    return txt



def dofile(fn):
    """ Process a file. """

    global lineno, filename

    lchunks = []

    try:
        infile = file(fn, "r")
    except:
        error("Can't open '%s' for reading." % fn)
        
    lastfilename   = filename
    filename       = fn
    lastlineno     = lineno
    

    while 1:     # Read and process the file, line by line
        s = infile.readline()

        if not s:
            break

        lineno += 1

        # skip over comment lines ... these start with "## " (space is needed)
        if s.startswith("## "): continue

        while 1:         # loop needed for recursion
            blen=len(brackend)  # IMPORTANT, len can change so set here!

            e = s.find(brackend)
        
            if e == -1:  # no end macro, continue reading the file.
                break

            # now we need to find the matching '[['. We check back from the ']]'
            # and prepend previously processed stuff til we find it (or the file start)
            
            lln = lineno

            while 1:
                b = s.rfind(brackstart, 0, e)

                if b > -1:
                    lineno = lln
                    rp = parse( s[b+blen:e] )
                    s = s[:b] + rp + s[e+blen:]
                    break

                if not lchunks: 
                    error("No '%s' found." % brackstart)
                
                s = lchunks.pop(-1) + s
                e = s.find(brackend)
                lineno =- 1

        lchunks.append(s)

    lineno = lastlineno
    filename = lastfilename

    return lchunks


############### mainline program  ##########


infile,outpath = opts()

lines=[]

if env['hook_start']:
    lines.append( env['hook_start']() )

lines.extend( dofile(infile) )

if env['hook_eof']:
    lines.append( env['hook_eof']() )

for a in lines:
    if a and len(a.strip()):  # this DOES NOTE get rid of all blank lines!
        outpath.write(a)

